#!/opt/wapt/python
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

# old function to install waptserver on windows. need to be rewritten (switch to nginx, websocket, etc.)

import shutil
import os
import sys
from win32api import GetUserName

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

from waptserver.config import __version__

from optparse import OptionParser
import logging
import subprocess
import setuphelpers
import datetime

import jinja2
import time
import random
import string
import base64
import uuid
import socket
import traceback

from setuphelpers import run
from waptutils import setloglevel, ensure_unicode
from waptcrypto import SSLPrivateKey, SSLCertificate, SSLCRL
from passlib.hash import pbkdf2_sha256

import waptserver.config
from waptserver.utils import logger, mkdir_p

from waptpackage import WaptLocalRepo
from waptserver.model import load_db_config, Packages


def fqdn():
    """ Guess FQDN from hostname - return default value if nothing found"""
    try:
        fqdn = socket.getfqdn()
        if not fqdn:
            fqdn = 'wapt'
        if '.' not in fqdn:
            fqdn += '.lan'
    except:
        fqdn = 'srvwapt'
    return fqdn


def create_dhparam(key_size=2048):
    logger.info('Create dhparam size %s. This can take a while...' % key_size)
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import dh
    parameters = dh.generate_parameters(generator=2, key_size=key_size, backend=default_backend())
    return parameters.parameter_bytes(serialization.Encoding.PEM, format=serialization.ParameterFormat.PKCS3)


def install_windows_nssm_service(
        service_name, service_binary, service_parameters, service_logfile, service_dependencies=None):
    """Setup a program as a windows Service managed by nssm
    >>> install_windows_nssm_service("WAPTServer",
        os.path.abspath(os.path.join(wapt_root_dir,'waptpython.exe')),
        os.path.abspath(__file__),
        os.path.join(log_directory,'nssm_waptserver.log'),
        service_logfile,
        'WAPTApache')
    """
    import setuphelpers
    from setuphelpers import registry_set, REG_DWORD, REG_EXPAND_SZ, REG_MULTI_SZ, REG_SZ
    datatypes = {
        'dword': REG_DWORD,
        'sz': REG_SZ,
        'expand_sz': REG_EXPAND_SZ,
        'multi_sz': REG_MULTI_SZ,
    }

    if setuphelpers.service_installed(service_name):
        if not setuphelpers.service_is_stopped(service_name):
            logger.info('Stop running "%s"' % service_name)
            setuphelpers.run('net stop "%s" /yes' % service_name)
            while not setuphelpers.service_is_stopped(service_name):
                logger.debug('Waiting for "%s" to terminate' % service_name)
                time.sleep(2)

        logger.info('Unregister existing "%s"' % service_name)
        setuphelpers.run('sc delete "%s"' % service_name)

    if not setuphelpers.iswin64():
        raise Exception('Windows 32bit install not supported')

    nssm = os.path.join(wapt_root_dir, 'waptservice', 'win64', 'nssm.exe')

    logger.info('Register service "%s" with nssm' % service_name)
    cmd = '"{nssm}" install "{service_name}" "{service_binary}" {service_parameters}'.format(
        nssm=nssm,
        service_name=service_name,
        service_binary=service_binary,
        service_parameters=service_parameters
    )
    logger.info('running command : %s' % cmd)
    setuphelpers.run(cmd)

    # fix some parameters (quotes for path with spaces...
    params = {
        'Description': 'sz:%s' % service_name,
        'DelayedAutostart': 1,
        'DisplayName': 'sz:%s' % service_name,
        'AppStdout': r'expand_sz:{}'.format(service_logfile),
        'ObjectName': r'NT AUTHORITY\NetworkService',
        'Parameters\\AppStderr': r'expand_sz:{}'.format(service_logfile),
        'Parameters\\AppParameters': r'expand_sz:{}'.format(service_parameters),
        'Parameters\\AppNoConsole': 1,
    }

    root = setuphelpers.HKEY_LOCAL_MACHINE
    base = r'SYSTEM\CurrentControlSet\services\%s' % service_name
    for key in params:
        if isinstance(params[key], int):
            (valuetype, value) = ('dword', params[key])
        elif ':' in params[key]:
            (valuetype, value) = params[key].split(':', 1)
            if valuetype == 'dword':
                value = int(value)
        else:
            (valuetype, value) = ('sz', params[key])
        fullpath = base + '\\' + key
        (path, keyname) = fullpath.rsplit('\\', 1)
        if keyname == '@' or keyname == '':
            keyname = None
        registry_set(root, path, keyname, value, type=datatypes[valuetype])

    if service_dependencies:
        logger.info(
            'Register dependencies for service "%s" with nssm : %s ' %
            (service_name, service_dependencies))
        cmd = '"{nssm}" set "{service_name}" DependOnService {service_dependencies}'.format(
            nssm=nssm,
            service_name=service_name,
            service_dependencies=service_dependencies
        )
        logger.info('running command : %s' % cmd)
        setuphelpers.run(cmd)

        # fullpath = base+'\\' + 'DependOnService'
        #(path,keyname) = fullpath.rsplit('\\',1)
        # registry_set(root,path,keyname,service_dependencies,REG_MULTI_SZ)


def make_nginx_config(wapt_root_dir, conf, force=False):
    """Create a nginx default config file to server wapt_folder and reverse proxy waptserver
    Create a key and self signed certificate.

    Args:
        wapt_root_dir (str)
        conf (dict) : waptserver config file

    Returns:
        str: path to nginx conf file
    """
    ap_conf_dir = os.path.join(
        wapt_root_dir,
        'waptserver',
        'nginx',
        'conf')
    ap_file_name = 'nginx.conf'
    ap_conf_file = os.path.join(ap_conf_dir, ap_file_name)
    ap_ssl_dir = os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'ssl')

    ap_ssl_dhparam_file = os.path.join(ap_ssl_dir, 'dhparam.pem')
    if not (os.path.isdir(ap_ssl_dir)):
        os.makedirs(ap_ssl_dir)
    if not(os.path.isfile(ap_ssl_dhparam_file)) or os.stat(ap_ssl_dhparam_file).st_size==0:
        with open(ap_ssl_dhparam_file, 'wb') as f:
            f.write(create_dhparam(dhparam_key_size))

    if os.path.isfile(ap_conf_file) and not force:
        if 'waptserver' in open(ap_conf_file, 'r').read():
            return ap_conf_file

    setuphelpers.mkdirs(ap_ssl_dir)

    key_fn = os.path.join(ap_ssl_dir, 'key.pem')
    key = SSLPrivateKey(key_fn)
    if not os.path.isfile(key_fn):
        logger.info('Create SSL RSA Key %s' % key_fn)
        key.create()
        key.save_as_pem()

    cert_fn = os.path.join(ap_ssl_dir, 'cert.pem')
    if os.path.isfile(cert_fn):
        crt = SSLCertificate(cert_fn)
        if crt.cn != fqdn():
            shutil.move(cert_fn, "%s-%s.old" % (cert_fn, '{:%Y%m%d-%Hh%Mm%Ss}'.format(datetime.datetime.now())))
            crt = key.build_sign_certificate(cn=fqdn(), dnsname=fqdn(), is_code_signing=False)
            logger.info('Create X509 cert %s' % cert_fn)
            crt.save_as_pem(cert_fn)
    else:
        crt = key.build_sign_certificate(cn=fqdn(), dnsname=fqdn(), is_code_signing=False)
        logger.info('Create X509 cert %s' % cert_fn)
        crt.save_as_pem(cert_fn)

    # write config file
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.join(wapt_root_dir, 'waptserver', 'scripts')))
    template = jinja_env.get_template('waptwindows.nginxconfig.j2')
    template_variables = {
        'wapt_repository_path': os.path.dirname(conf['wapt_folder']).replace('\\', '/'),
        'wapt_folder': os.path.dirname(conf['wapt_folder']).replace('\\', '/'),
        'waptserver_port': conf['waptserver_port'],
        'windows': True,
        'ssl': True,
        'force_https': False,
        'use_kerberos': False,
        'wapt_ssl_key_file': key_fn.replace('\\', '/'),
        'wapt_ssl_cert_file': cert_fn.replace('\\', '/'),
        'log_dir': os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'logs').replace('\\', '/'),
        'wapt_root_dir': wapt_root_dir.replace('\\', '/'),
        'nginx_http': conf['nginx_http'],
        'nginx_https': conf['nginx_https'],
        'use_ssl_client_auth': conf.get('use_ssl_client_auth', False),
        'clients_signing_certificate': conf.get('clients_signing_certificate') and conf.get('clients_signing_certificate').replace('\\', '/'),
        'known_certificates_folder': conf.get('known_certificates_folder', None) and conf.get('known_certificates_folder', None).replace('\\', '/'),
        'clients_signing_crl': conf.get('clients_signing_crl', None) and conf.get('clients_signing_crl', None).replace('\\', '/'),
        'htpasswd_path': conf.get('htpasswd_path', None) and conf.get('htpasswd_path', None).replace('\\', '/'),
        'wapt_dhparam_file': ap_ssl_dhparam_file.replace('\\', '/'),
        'has_dhparam': os.path.isfile(ap_ssl_dhparam_file),
    }

    config_string = template.render(template_variables)
    logger.info('Create nginx conf file %s' % ap_conf_file)
    with open(ap_conf_file, 'wt') as dst_file:
        dst_file.write(config_string)
    return ap_conf_file


def install_nginx_service(options, conf=None):
    if conf is None:
        conf = waptserver.config.load_config(options.configfile)

    logger.info("Register nginx frontend")
    repository_path = os.path.join(wapt_root_dir, 'waptserver', 'repository')
    for repo_path in ('.', 'wapt', 'wapt-host', 'waptwua'):
        mkdir_p(os.path.join(repository_path, repo_path))
        run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % os.path.join(repository_path, repo_path))
    mkdir_p(os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'temp'))
    run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'temp')))

    run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'logs'))
    run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % os.path.join(wapt_root_dir, 'log'))

    make_nginx_config(wapt_root_dir, conf, force=options.force)
    service_binary = os.path.abspath(os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'nginx.exe'))
    service_parameters = ''
    service_logfile = os.path.join(log_directory, 'nssm_nginx.log')

    service_name = 'WAPTNginx'
    if setuphelpers.service_installed(service_name) and setuphelpers.service_is_running(service_name):
        setuphelpers.service_stop(service_name)
    #logger.info('Register "%s" in registry' % service_name)
    install_windows_nssm_service(service_name, service_binary, service_parameters, service_logfile)
    time.sleep(5)
    if setuphelpers.service_installed(service_name) and not setuphelpers.service_is_running(service_name):
        setuphelpers.service_start(service_name)


def migrate_pg_db(old_pgsql_root_dir, old_pgsql_data_dir, pgsql_root_dir, pgsql_data_dir):
    #
    try:
        cwd = os.getcwd()
        tmpdir = os.path.join(cwd, 'temp')
        mkdir_p(tmpdir)
        os.chdir(tmpdir)
        setuphelpers.run(r'icacls "%s" /t /grant  "*S-1-1-0":(OI)(CI)(M)' % tmpdir)

        old_pgsql_root_dir = old_pgsql_root_dir.replace('\\', '/')
        old_pgsql_data_dir = old_pgsql_data_dir.replace('\\', '/')
        pgsql_root_dir = pgsql_root_dir.replace('\\', '/')
        pgsql_data_dir = pgsql_data_dir.replace('\\', '/')

        cmd = r'"{pgsql_root_dir}/bin/pg_upgrade.exe" -U postgres --old-datadir "{old_pgsql_data_dir}" --new-datadir "{pgsql_data_dir}" --old-bindir "{old_pgsql_root_dir}/bin" --new-bindir "{pgsql_root_dir}/bin"'.format(**locals())

        logger.info('Running %s' % cmd)
        logger.info(run(cmd, cwd=tmpdir))
        shutil.move(old_pgsql_root_dir, old_pgsql_root_dir+'.old')
        shutil.move(old_pgsql_data_dir, old_pgsql_data_dir+'.old')
        return True
    except Exception as e:
        logger.info('Unable to migrate database : %s' % ensure_unicode(e))
        return False
    finally:
        os.chdir(cwd)


def configure_max_connections(pg_sql_data_dir):
    config_file_path = os.path.join(pg_sql_data_dir, 'postgresql.conf')
    with open(config_file_path, "r+") as config_file:
        new_lines = []
        found = False
        for line in config_file.readlines():
            if 'max_connections' in line:
                new_lines.append('max_connections = 1000\n')
                found = True
            else:
                new_lines.append(line)
        if not(found):
            new_lines.append('max_connections = 1000\n')
        config_file.seek(0)
        config_file.truncate()
        config_file.writelines(new_lines)


def install_postgresql_service(options, conf=None):
    if conf is None:
        conf = waptserver.config.load_config(options.configfile)
    logger.info("Install postgres database")

    pgsql_root_dir = r'%s\waptserver\pgsql-9.6' % wapt_root_dir
    pgsql_data_dir = r'%s\waptserver\pgsql_data-9.6' % wapt_root_dir
    pgsql_data_dir = pgsql_data_dir.replace('\\', '/')

    logger.info("About to build database directory")
    if setuphelpers.service_installed('waptserver') and setuphelpers.service_is_running('waptserver'):
        logger.info('Stopping waptserver')
        setuphelpers.service_stop('waptserver')
        # waiting for postgres to be ready
        time.sleep(1)

    if setuphelpers.service_installed('wapttasks') and setuphelpers.service_is_running('wapttasks'):
        logger.info('Stopping wapttasks')
        setuphelpers.service_stop('wapttasks')
        # waiting for postgres to be ready
        time.sleep(1)

    if setuphelpers.service_installed('waptpostgresql') and setuphelpers.service_is_running('waptpostgresql'):
        logger.info('Stopping postgresql')
        setuphelpers.service_stop('waptpostgresql')
        # waiting for postgres to be ready
        time.sleep(2)

    if not os.path.exists(os.path.join(pgsql_data_dir, 'postgresql.conf')):
        setuphelpers.mkdirs(pgsql_data_dir)

        # need to have specific write acls for current user otherwise initdb fails...
        setuphelpers.run(r'icacls "%s" /t /grant  "%s":(OI)(CI)(M)' % (pgsql_data_dir, GetUserName()))

        setuphelpers.run(r'"%s\bin\initdb" -U postgres -E=UTF8 -D "%s"' % (pgsql_root_dir, pgsql_data_dir))
        setuphelpers.run(r'icacls "%s" /t /grant  "*S-1-5-20":(OI)(CI)(M)' % pgsql_data_dir)
    else:
        logger.info("Database already instanciated, doing nothing")

    logger.info("Start postgresql database")

    if setuphelpers.service_installed('WaptPostgresql'):
        if setuphelpers.service_is_running('WaptPostgresql'):
            setuphelpers.service_stop('waptPostgresql')
        setuphelpers.service_delete('waptPostgresql')

    cmd = r'"%s\bin\pg_ctl" register -N WAPTPostgresql -U "nt authority\networkservice" -S auto -D "%s"  ' % (pgsql_root_dir, pgsql_data_dir)
    run(cmd)
    setuphelpers.run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % log_directory)
    setuphelpers.run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % pgsql_data_dir)

    # try to migrate from old version (pg 9.4, wapt 1.5)
    old_pgsql_root_dir = r'%s\waptserver\pgsql' % wapt_root_dir
    old_pgsql_data_dir = r'%s\waptserver\pgsql_data' % wapt_root_dir
    old_pgsql_data_dir = old_pgsql_data_dir.replace('\\', '/')

    if os.path.isdir(old_pgsql_data_dir) and os.path.isdir(old_pgsql_root_dir):
        logger.info('Migrating database from previous postgresql DB')
        migrate_pg_db(old_pgsql_root_dir, old_pgsql_data_dir, pgsql_root_dir, pgsql_data_dir)

    configure_max_connections(pgsql_data_dir)

    logger.info('Starting postgresql')
    if not setuphelpers.service_is_running('waptpostgresql'):
        setuphelpers.service_start('waptpostgresql')
        # waiting for postgres to be ready
        time.sleep(2)

    logger.info("Checking wapt database")
    import psycopg2
    from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
    conn = None
    cur = None
    try:
        conn = psycopg2.connect('dbname=template1 user=postgres')
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cur = conn.cursor()
        cur.execute("select 1 from pg_roles where rolname='%(db_user)s'" % conf)
        val = cur.fetchone()
        if val is None:
            logger.info("%(db_user)s pgsql user does not exists, creating %(db_user)s user" % conf)
            cur.execute("create user %(db_user)s" % conf)

        cur.execute("select 1 from pg_database where datname='%(db_name)s'" % conf)
        val = cur.fetchone()
        if val is None:
            logger.info("Database %(db_name)s does not exists, creating %(db_name)s db" % conf)
            cur.execute("create database %(db_name)s owner %(db_user)s" % conf)

    finally:
        if cur:
            cur.close()
        if conn:
            conn.close()

    logger.info("Creating/upgrading wapt db tables")
    run(r'"%s\waptpython.exe" -I "%s\waptserver\model.py" init_db -c "%s"' % (wapt_root_dir, wapt_root_dir, options.configfile))
    logger.info("Done")

    logger.info('Import local Packages data into database')

    repo = WaptLocalRepo(conf['wapt_folder'])
    load_db_config(conf)
    Packages.update_from_repo(repo)


def install_waptserver_service(options, conf=None):
    if setuphelpers.service_installed('WAPTServer'):
        if setuphelpers.service_is_running('WAPTServer'):
            setuphelpers.service_stop('WAPTServer')
        setuphelpers.service_delete('WAPTServer')

    if conf is None:
        conf = waptserver.config.load_config(options.configfile)

    conf_dir = os.path.join(wapt_root_dir, 'conf')
    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)

    if not conf.get('server_uuid'):
        conf['server_uuid'] = str(uuid.uuid1())
        waptserver.config.write_config_file(options.configfile, conf)

    logger.info("Install waptserver")
    service_binary = os.path.abspath(os.path.join(wapt_root_dir, 'waptpython.exe'))
    service_parameters = '"%s"' % os.path.join(wapt_root_dir, 'waptserver', 'server.py')
    service_logfile = os.path.join(log_directory, 'nssm_waptserver.log')
    service_dependencies = 'WAPTPostgresql'
    install_windows_nssm_service('WAPTServer', service_binary, service_parameters, service_logfile, service_dependencies)

    tasks_db = os.path.join(wapt_root_dir, 'db')
    mkdir_p(tasks_db)

    if not conf.get('secret_key'):
        conf['secret_key'] = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(64))
        waptserver.config.write_config_file(options.configfile, conf)

    if options.setpassword:
        conf['wapt_password'] = pbkdf2_sha256.hash(base64.b64decode(options.setpassword))
        waptserver.config.write_config_file(options.configfile, conf)

    clients_signing_certificate = conf.get('clients_signing_certificate')
    clients_signing_key = conf.get('clients_signing_key')
    clients_signing_crl = conf.get('clients_signing_crl')

    if not clients_signing_certificate or not clients_signing_key:
        clients_signing_certificate = os.path.join(wapt_root_dir, 'conf', 'ca-%s.crt' % fqdn())
        clients_signing_key = os.path.join(wapt_root_dir, 'conf', 'ca-%s.pem' % fqdn())

        conf['clients_signing_certificate'] = clients_signing_certificate
        conf['clients_signing_key'] = clients_signing_key
        waptserver.config.write_config_file(options.configfile, conf)

    if clients_signing_certificate is not None and clients_signing_key is not None and not os.path.isfile(clients_signing_certificate):
        logger.info('Create a certificate and key for clients certificate signing')

        key = SSLPrivateKey(clients_signing_key)
        if not os.path.isfile(clients_signing_key):
            logger.info('Create SSL RSA Key %s' % clients_signing_key)
            key.create()
            key.save_as_pem()

        crt = key.build_sign_certificate(cn=fqdn(), is_code_signing=False, is_ca=True)
        logger.info('Create X509 cert %s' % clients_signing_certificate)
        crt.save_as_pem(clients_signing_certificate)

    # known certificates
    ssl_dir = conf['known_certificates_folder']
    if not os.path.isdir(ssl_dir):
        # move existing ssl dir in wapt repo to parent dir (default location)
        if os.path.isdir(os.path.join(conf['wapt_folder'], 'ssl')):
            shutil.move(os.path.join(conf['wapt_folder'], 'ssl'), ssl_dir)
        else:
            os.makedirs(ssl_dir)

    if clients_signing_certificate is not None and clients_signing_key is not None and clients_signing_crl is not None and not os.path.isfile(clients_signing_crl):
        logger.info('Create a CRL for clients certificate signing')
        key = SSLPrivateKey(clients_signing_key)
        crt = SSLCertificate(clients_signing_certificate)
        crl = SSLCRL(clients_signing_crl, cacert=crt, cakey=key)
        crl.revoke_cert()    # ensure Packages index
        crl.save_as_pem()

    repo = WaptLocalRepo(conf['wapt_folder'])
    repo.update_packages_index()

    # Migrate file for new version waptwua
    wuafolder = conf['waptwua_folder']
    for (root, dirs, files) in list(os.walk(wuafolder, topdown=False)):
        if root == os.path.join(wuafolder, '.stfolder'):
            continue
        for f in files:
            oldpath = os.path.join(root, f)
            newpath = os.path.join(wuafolder, f)
            if os.path.isfile(newpath):
                continue
            logger.info('Move %s --> %s' % (oldpath, newpath))
            shutil.move(oldpath, newpath)
        for d in dirs:
            if d == '.stfolder':
                continue
            logger.info('Delete folder %s' % os.path.join(root, d))
            shutil.rmtree(os.path.join(root, d))

    run('icacls "%s" /t /grant "*S-1-5-20":(OI)(CI)(M) /inheritancelevel:e' % (wapt_root_dir))

    if setuphelpers.service_installed('WAPTServer'):
        if not setuphelpers.service_is_running('WAPTServer'):
            setuphelpers.service_start('WAPTServer')


def install_wapttasks_service(options, conf=None):
    if setuphelpers.service_installed('WAPTTasks'):
        if setuphelpers.service_is_running('WAPTTasks'):
            setuphelpers.service_stop('WAPTTasks')
        setuphelpers.service_delete('WAPTTasks')

    if conf is None:
        conf = waptserver.config.load_config(options.configfile)
    logger.info("Install wapttasks")
    service_binary = os.path.abspath(os.path.join(wapt_root_dir, 'waptpython.exe'))
    service_parameters = '"%s" %s' % (os.path.join(wapt_root_dir, 'waptserver', 'wapthuey.py'), 'waptserver.tasks_common.huey -f -w 2')
    service_logfile = os.path.join(log_directory, 'nssm_wapttasks.log')
    service_dependencies = 'WAPTPostgresql'
    install_windows_nssm_service('WAPTTasks', service_binary, service_parameters, service_logfile, service_dependencies)

    tasks_db = os.path.join(wapt_root_dir, 'db')
    setuphelpers.run(r'icacls "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % tasks_db)

    if setuphelpers.service_installed('WAPTTasks'):
        if not setuphelpers.service_is_running('WAPTTasks'):
            setuphelpers.service_start('WAPTTasks')


if __name__ == '__main__':
    usage = """\
    %prog [-c configfile] [install_nginx install_postgresql install_waptserver] [--dhparam-key-size=SIZE]

    WAPT Server services setup.

    actions is either :
      <nothing> : run service in foreground
      install   : install as a Windows service managed by nssm
      uninstall : uninstall Windows service managed by nssm

    """
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    setloglevel(logger,'info')

    config_filename = os.path.join(wapt_root_dir, 'conf', 'waptserver.ini')

    parser = OptionParser(usage=usage, version='winsetup.py ' + __version__)
    parser.add_option('-c', '--config', dest='configfile', default=config_filename,
                      help='Config file full path (default: %default)')

    parser.add_option('-l', '--loglevel', dest='loglevel', default=None, type='choice',
                      choices=['debug',   'warning', 'info', 'error', 'critical'],
                      metavar='LOGLEVEL', help='Loglevel (default: info)')
    parser.add_option('-d', '--devel', dest='devel', default=False, action='store_true',
                      help='Enable debug mode (for development only)')
    parser.add_option('-f', '--force', dest='force', default=False, action='store_true',
                      help='Force rewrite nginx config')
    parser.add_option('-p', '--setpassword', dest='setpassword', default=None,
                      help='Set wapt server admin password. Value must be encoded in base64 (default: %default)')
    parser.add_option(
        '--dhparam-key-size',
        dest='dhparam_key_size',
        default=1024,
        metavar='NUMBER',
        type='int',
        help='Size for dhparam key')

    (options, args) = parser.parse_args()

    if options.loglevel is not None:
        setloglevel(logger, options.loglevel)

    conf = waptserver.config.load_config(options.configfile)

    if conf['wapt_folder'].endswith('\\') or conf['wapt_folder'].endswith('/'):
        conf['wapt_folder'] = conf['wapt_folder'][:-1]

    log_directory = os.path.join(wapt_root_dir, 'log')
    if not os.path.exists(log_directory):
        os.mkdir(log_directory)
        run(r'icacls "%s" /t /grant  "*S-1-5-20":(OI)(CI)(M)' % os.path.join(wapt_root_dir, 'log'))

    h = logging.FileHandler(filename=os.path.join(log_directory,'waptserver_install.log'))
    h.setFormatter(formatter)
    logger.addHandler(h)

    dhparam_key_size = options.dhparam_key_size

    if args == ['all']:
        args = ['install_nginx', 'install_postgresql', 'install_waptserver']

    try:
        for action in args:
            if action == 'install_nginx':
                logger.info('Installing nginx as a service managed by nssm')
                install_nginx_service(options, conf)
            elif action == 'install_postgresql':
                logger.info('Installing PostgreSQL as a service managed by nssm')
                install_postgresql_service(options, conf)
            elif action == 'install_waptserver':
                logger.info('Installing WAPT Server as a service managed by nssm')
                install_waptserver_service(options, conf)
                install_wapttasks_service(options, conf)
            elif action == 'create-check-dhparam':
                logger.info('Create or check dhparam file')
                ap_ssl_dir = os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'ssl')
                ap_dhparam = os.path.join(ap_ssl_dir, 'dhparam.pem')
                if not (os.path.isdir(ap_ssl_dir)):
                    os.makedirs(ap_ssl_dir)
                if not(os.path.isfile(ap_dhparam)):
                    with open(ap_dhparam, 'wb') as f:
                        f.write(create_dhparam(dhparam_key_size))
    except Exception as e:
        logger.critical(traceback.format_exc())
        raise
