#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys

from waptserver.config import __version__

__all__ = [
    'trigger_download_wsusscan',
    'wsusscan2_status',
    'wsus_downloads',
    'wsus_downloads_delete',
    'windows_updates',
    'windows_updates_delete',
    'download_windows_updates',
    'windows_products',
    'windows_updates_classifications',
    'trigger_delete_non_assigned_kb',
    'get_storage_used_by_kbs',
]

import email.utils

from flask import request, send_from_directory, Blueprint
from werkzeug.utils import secure_filename

import hashlib
import json
import ujson
import random
import re
import requests
import shutil
import stat
import subprocess
import time
import traceback
import urllib.parse
import uuid

import waptserver.config
from waptutils import wget, datetime2isodate, makepath

from waptserver.utils import *
from waptserver.model import *
from waptserver.config import get_http_proxies

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, require_wapt_db, requires_enterprise

from waptserver.wsus_tasks import download_wsusscan, download_windows_update_task, download_missing_cabs, huey,authorized_url_wsus,get_update_local_path, delete_non_assigned_kb
from waptserver.wsus_common import simplematch, products_id, update_classifications_id, get_product_id
from waptserver.wsus_common import check_sha1_filename, wuredist_extract_and_fetch, cabextract, check_cab

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

wsus = Blueprint('wsus', __name__)

DEFAULT_CONFIG_FILE = os.path.join(wapt_root_dir, 'conf', 'waptserver.ini')
config_file = DEFAULT_CONFIG_FILE

conf = waptserver.config.load_config(config_file)


@wsus.route('/api/v3/download_wsusscan')
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def trigger_download_wsusscan():
    force = bool(int(request.args.get('force', 0)))

    in_progress = WsusDownloadTasks.select(WsusDownloadTasks.id, WsusDownloadTasks.started_on, WsusDownloadTasks.msg).where(
        (WsusDownloadTasks.kind == 'WSUSCAN') &
        (~WsusDownloadTasks.status.in_(['OK', 'ERROR']))
    )

    if list(in_progress):
        result = list(in_progress)
        msg = 'A download task is already in progress: started on %s, %s' % (','.join([t.started_on for t in in_progress]), ','.join([t.msg for t in in_progress]))
    else:
        logger.info('Triggering download_wsusscan with parameter ' + str(force))
        result = download_wsusscan(force=force)
        msg = 'new wsusscan2 download task queued'

    return make_response(result=result, msg=msg)


@wsus.route('/api/v3/download_missing_cabs')
@requires_enterprise
@requires_auth(['admin', 'edit_wsus_package'])
def trigger_download_missing_cabs():
    result = download_missing_cabs()
    return make_response(result=result, msg='Download of missing cabs enqueued')

@wsus.route('/api/v3/show_download_missing_cabs')
@requires_enterprise
@requires_auth(['admin', 'edit_wsus_package'])
def show_trigger_download_missing_cabs():
    with WaptDB():
        updates = (HostWsus.select(WsusUpdates.update_id, WsusUpdates.download_urls)
                   .join(WsusUpdates).distinct()
                   .where((HostWsus.installed == False) & (HostWsus.allowed == True)))
        result = []
        for update in updates:
            for url in update.update_id.download_urls:
                if not authorized_url_wsus(url):
                    continue
                if authorized_url_wsus(url):
                    local_path = get_update_local_path(url)
                    if not os.path.isfile(local_path) and not url in result:
                        result.append(url)

    return make_response(result=result, msg='Missing cabs')



@wsus.route('/api/v2/wsusscan2_status')
@requires_enterprise
@requires_auth(['admin', 'view'])
def wsusscan2_status():
    wsus_filename = os.path.join(conf['waptwua_folder'], 'wsusscn2.cab')
    tmp_filename = os.path.join(conf['waptwua_folder'], 'wsusscn2.cab.tmp')

    success = False
    downloading = False
    exc_type = 'unknown'
    data = {}

    try:
        stats = os.stat(wsus_filename)
        success = True
        data.update({
            'wsus_timestamp': stats[stat.ST_MTIME],
            'wsus_size':      stats[stat.ST_SIZE],
        })
    except Exception as e:
        exc_type = type(e).__name__.lower()

    try:
        tmp_stats = os.stat(tmp_filename)
        downloading = True
        data.update({
            'tmp_wsus_timestamp': tmp_stats[stat.ST_MTIME],
            'tmp_wsus_size':      tmp_stats[stat.ST_SIZE],
        })
    except Exception:
        pass

    if success:
        msg = 'wsusscn2.cab present'
        error_code = ''
    else:
        if downloading:
            msg = 'wsusscn2.cab absent, is another download running?'
        else:
            msg = 'wsusscn2.cab absent, please download it first'
        error_code = exc_type

    return make_response(success=success, result=data, msg=msg, error_code=error_code)


@wsus.route('/api/v3/wsus_downloads', methods=['GET', 'POST', 'DELETE'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def wsus_downloads():
    """GET or DELETE the history of WSUS download tasks

    Args:
        kind
        ids
        status
        skipped
        limit

    Returns:

    """

    try:
        if request.method in ['GET', 'POST']:
            data = []
            filter = {}

            query = WsusDownloadTasks.select()
            filter = []

            if 'kind' in request.args:
                filter.append(WsusDownloadTasks.kind.in_(ensure_list(request.args['kind'])))
            if 'ids' in request.args:
                filter.append(WsusDownloadTasks.id.in_([int(s) for s in ensure_list(request.args['ids'])]))
            if 'status' in request.args:
                filter.append(WsusDownloadTasks.status.in_(ensure_list(request.args['status'])))
            if 'skipped' in request.args:
                skipped = bool(int(request.args['skipped']))
                filter.append(WsusDownloadTasks.skipped == skipped)

            if filter:
                where = filter[0]
                for q in filter[1:]:
                    where = where & q
                query = query.where(where)

            query = query.order_by(WsusDownloadTasks.created_on.desc())

            limit = int(request.args.get('limit', '0'))
            if limit:
                query = query.limit(limit)
            for log in query.dicts():
                data.append(log)
            return make_response(result=data)
        elif request.method == 'DELETE':
            ids = [int(i) for i in ensure_list(request.args['ids'])]
            result = WsusDownloadTasks.delete().where(WsusDownloadTasks.id.in_(ids)).execute()
            return make_response(result=result, msg='%s tasks deleted' % result)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('wsus_downloads failed %s' % (repr(e)))
        return make_response_from_exception(e)


@wsus.route('/api/v3/wsus_downloads_delete', methods=['POST'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def wsus_downloads_delete():
    start_time = time.time()
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("wsus_downloads_delete: No ids supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            deleted_count = WsusDownloadTasks.delete().where(WsusDownloadTasks.id.in_(id_list)).execute()
            trans.commit()
            return make_response(deleted_count, success=True, request_time=time.time() - start_time)
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@wsus.route('/api/v3/windows_products')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def windows_products():
    result = list(WsusUpdates.raw('select product,count(*) as wu_count from wsusupdates group by product').dicts())
    return make_response(msg=_('Windows Products'), result=result)


@wsus.route('/api/v3/windows_updates_classifications')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def windows_updates_classifications():
    result = sorted([r[0] for r in WsusUpdates.select(fn.DISTINCT(WsusUpdates.classification)).tuples()])
    return make_response(msg=_('Win updates classifications'), result=result)


@wsus.route('/api/v3/windows_updates')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def windows_updates():
    """
    """
    classifications = ensure_list(request.args.get('classifications'))
    products = ensure_list(request.args.get('products'))
    filter = []
    if classifications:
        filter.append(WsusUpdates.classification.in_(classifications))
    if products:
        filter.append(WsusUpdates.product.in_(products))

    w = None
    for f in filter:
        if w is None:
            w = f
        else:
            f = f & w

    result = list(WsusUpdates.select().where(w).order_by(WsusUpdates.created_on.desc()).dicts())

    return make_response(msg=_('Windows Updates'), result=result)


@wsus.route('/api/v3/windows_updates_delete', methods=['POST'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def windows_updates_delete():
    start_time = time.time()
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("windows_updates_delete: No ids supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            all_urls = WsusUpdates.select(WsusUpdates.download_urls).where(WsusUpdates.update_id.in_(id_list)).tuples()
            for urls in all_urls:
                for url in urls[0]:
                    filename = urllib.parse.urlparse(url).path.split('/')[-1]
                    target = os.path.join(conf['waptwua_folder'], filename)
                    if os.path.isfile(target):
                        os.unlink(target)

            deleted_count = WsusUpdates.update(downloaded_on=None).where(WsusUpdates.update_id.in_(id_list)).execute()
            trans.commit()
            return make_response(deleted_count, success=True, request_time=time.time() - start_time)
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@wsus.route('/api/v3/hosts_for_windows_update')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def hosts_for_windows_update():
    """Returns list of hosts requiring the supplied windows update

    Args:
        update_id (str)

    Returns:
        list of Hosts

    """
    update_id = request.args.get('update_id')
    result = list(HostWsus.select().where(HostWsus.host == update_id).order_by(HostWsus.created_on.desc()).dicts())

    return make_response(msg=_('Hosts for Windows Update %s') % (update_id,), result=result)


@wsus.route('/api/v3/download_windows_update', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def download_windows_updates():
    """Add download tasks for the given list of update_id

    """
    try:
        if request.is_json:
            update_ids = request.json.get('update_ids')
        else:
            update_ids = ensure_list(request.args.get('update_ids'), allow_none=True)

        force = bool(int(request.args.get('force', 0)))

        updates = WsusUpdates.select(WsusUpdates.update_id, WsusUpdates.download_urls, WsusUpdates.max_download_size, WsusUpdates.downloaded_on)
        if update_ids is not None:
            updates = updates.where(WsusUpdates.update_id.in_(update_ids))
        else:
            updates = updates.where(WsusUpdates.downloaded_on.is_null())

        urls = []
        for update in updates:
            for url in update.download_urls:
                filename = urllib.parse.urlparse(url).path.split('/')[-1]
                target = os.path.join(conf['waptwua_folder'], filename)

                if (force or not os.path.isfile(target)) and not url in urls:
                    already_queued = [t for t in huey.pending() if t.name == download_windows_update_task and t.data[0] == url]
                    if not already_queued:
                        urls.append(url)
                elif os.path.isfile(target) and not update.downloaded_on:
                    update.downloaded_on = datetime2isodate()
                    update.save()

        result = []
        for url in urls:
            task = download_windows_update_task(url, force)
            result.append(task.task.id)
            logger.info('Added a task %s to download update: %s' % (task.task.id, url))
        return make_response(msg=_('%s Windows updates tasks enqueued') % (len(result),), result=result)
    except Exception as e:
        return make_response_from_exception(e)


@wsus.route('/waptwua/<path:wsuspackage>')
@requires_enterprise
def get_wua_package(wsuspackage):
    fileparts = wsuspackage.split('/')
    full_path = os.path.join(conf['waptwua_folder'], *fileparts[:-1])
    package_name = secure_filename(fileparts[-1])
    r = send_from_directory(full_path, package_name)
    if 'content-length' not in r.headers:
        r.headers.add_header('content-length', int(os.path.getsize(os.path.join(full_path, package_name))))
    return r


@wsus.route('/api/v2/download_wuredist')
@requires_enterprise
def download_wuredist():
    cab_url = 'http://update.microsoft.com/redist/wuredist.cab'
    wuredist_filename = os.path.join(conf['waptwua_folder'], 'wuredist.cab')
    tmp_filename = wuredist_filename + '.part'

    # should we remove the tmp filename when we get an error?
    cleanup_on_error = False

    force = request.args.get('force', False)
    dryrun = request.args.get('dryrun', False)

    stats = {}

    if os.path.isfile(tmp_filename):
        os.unlink(tmp_filename)

    try:
        http_proxy = get_http_proxies(conf)

        last_modified = requests.head(cab_url, proxies=http_proxy).headers['last-modified']

        new_cab_timestamp = float(email.utils.mktime_tz(email.utils.parsedate_tz(last_modified)))
        if os.path.isfile(wuredist_filename):
            current_cab_timestamp = os.stat(wuredist_filename).st_mtime
        else:
            current_cab_timestamp = 0.0

        logger.info('download_wuredist: current cab timestamp: %f, New cab timestamp: %f' % (current_cab_timestamp, new_cab_timestamp))

        if not os.path.isfile(wuredist_filename) or (new_cab_timestamp > current_cab_timestamp) or force:

            cleanup_on_error = True
            if dryrun:
                try:
                    if os.path.exists(tmp_filename):
                        os.unlink(tmp_filename)
                    os.link(wuredist_filename, tmp_filename)
                except Exception as e:
                    logger.error('download_wuredist: exception %s', str(e))
                    pass
            else:
                wget(cab_url, tmp_filename, proxies=http_proxy)

            file_stats = os.stat(tmp_filename)
            stats['file_timestamp'] = file_stats[stat.ST_MTIME]
            stats['file_size'] = file_stats[stat.ST_SIZE]
            stats['status'] = 'checking'
            # wuredist_history.save(stats)

            # TODO verify cryptographic signatures

            # check integrity
            cablist = check_cab(tmp_filename).splitlines()
            stats['cablist'] = cablist

            if os.path.isfile(wuredist_filename):
                os.unlink(wuredist_filename)
            shutil.move(tmp_filename, wuredist_filename)

            stats['status'] = 'parsing'
            # wuredist_history.save(stats)

        else:
            stats['skipped'] = True
            # wuredist_history.save(stats)

        tmpdir = os.path.join(conf['waptwua_folder'], 'wuredist.tmp')
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)
        mkdir_p(tmpdir)
        result = wuredist_extract_and_fetch(wuredist_filename, tmpdir, http_proxy)

        stats['status'] = 'finished'

    except Exception as e:
        stats['error'] = str(e)
        # wuredist_history.save(stats)

        if cleanup_on_error and os.path.isfile(tmp_filename):
            os.unlink(tmp_filename)

        logger.error("Error in download_wuredist: %s", str(e))
        logger.error('Trace:\n%s', traceback.format_exc())
        return make_response_from_exception(e)

    return make_response(msg=str(stats), success=True, result=result)


@wsus.route('/api/v3/cancel_download_tasks')
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
def cancel_download_tasks():
    """Cancel download tasks for the given list of task_id
    """
    task_ids = ensure_list(request.args.get('task_ids'))
    if not task_ids:
        result = huey.pending_count()
        huey.flush()
        msg = '%s pending download tasks flushed' % result
    else:
        for task_id in task_ids:
            huey.revoke_by_id(task_id)
        msg = '%s tasks revoked' % len(task_ids)
        result = len(task_ids)

    return make_response(msg=msg, result=result)

@wsus.route('/api/v3/trigger_delete_non_assigned_kb', methods=['GET'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
@require_wapt_db
def trigger_delete_non_assigned_kb():
    need_to_delete = str(request.args.get('to_delete'))
    off_the_grid = str(request.args.get('off_grid'))

    result = []
    result = delete_non_assigned_kb(need_to_delete, off_the_grid)

    return make_response(result=result)

@wsus.route('/api/v3/get_storage_used_by_kbs', methods=['GET'])
@requires_auth(['admin', 'edit_wsus_package'])
@requires_enterprise
def get_storage_used_by_kbs():

    language = str(request.args.get('Language'))

    storage = {}
    for row in glob.glob(makepath(conf['waptwua_folder'],'*')):
        storage[row] = os.stat(row).st_size

    data_storage = 0
    for files in storage:
        data_storage += storage[files]

    data_storage /= pow(1024, 2)
    if data_storage != 0.0:
        data_storage = '%.3f'%(data_storage)

    if language == 'fr':
        msg=f'Espace Utilisé par les MAJ Windows : {data_storage} Mo.'
    else:
        msg=f'Space used by Windows Updates : {data_storage} Mo.'

    return make_response(result=data_storage, msg=msg)

